/* -------------------------------------------------
                    mpx_8digit.c

   Softwaremodul zur Realisierung einer gemulti-
   plexten 8-stelligen 7-Segmentanzeige und zu-
   saetzlich 8 gemultiplexten Tastern.

   Modul ist nur verwendbar mit ATmega mit auf
   internem Takt eingestellten Fuses, da alle
   Portpins benoetigt werden (um die Leitungen des
   UART-Anschlusses frei zu halten).

   Es wird Timer1 fuer das Multiplexen verwendet

   MCU       : ATmega168 - 8 MHz Takt intern
               Fuses:  LO-> 0xE2  / HI-> 0xdf  / EX-> 0xF9
   IDE       : keine (Editor / make)
   Toolchain : AVR-GCC

   31.01.22 by R. Seelig
  -------------------------------------------------- */

#include "mpx_8digit.h"

volatile uint8_t  seg7_fb[8];
volatile uint8_t  key_stat;        // Status der Tastatur (jede Taste repraesentiert ein
                                   // Bit der uint8_t Variable

// mittels dig_pwm wird innerhalb des Interrupts eine Software-PWM realisiert, mittels
// derer eine Helligkeitssteuerung eines einzelnen Digits erfolgen kann. Zulaessige Werte
// fuer das Array sind 1..15. 1 entspricht volle Helligkeit, 15 entspricht sehr dunkel
volatile uint8_t  dig_pwm[8] = {1,1,1,1,1,1,1,1};


/* -------------------------------------------------------------------------------
 Bitmapmuster der Anzeige

       a            Bit-Nr.:  |  7  |  6  |  5  |  4  |  3  |  2  |  1  |  0  |
      ---           -----------------------------------------------------------
   f | g | b        Segment   |  dp |  g  |  f  |  e  |  d  |  c  |  b  |  a  |
      ---
   e |   | c
      ---   .
       d    dp

   Bsp. fuer "5" bei Display mit gemeinsamer Kathode

                    Bit-Nr.:  |  7  |  6  |  5  |  4  |  3  |  2  |  1  |  0  |
         ----------------------------------------------------------------------
                    Segment   |  dp |  g  |  f  |  e  |  d  |  c  |  b  |  a  |
         ----------------------------------------------------------------------
         leuchtendes Segment  |  0  |  1  |  1  |  0  |  1  |  1  |  0  |  1  | ==> 0x6d

   ------------------------------------------------------------------------------- */

uint8_t seg7_bmp[16] =
{
  0x3F, 0x06, 0x5B, 0x4F, 0x66, 0x6D, 0x7D, 0x07,
  0x7F, 0x6F, 0x77, 0x7C, 0x39, 0x5E, 0x79, 0x71 };


uint8_t seg7_dpmask = 0;            // Maske fuer anzuzeigende Dezimalpunkte
                                    // 0x00 = alle DPs aus, 0xFF = alle DPs an

/* --------------------------------------------------------
                         seg7_initall

     initialisiert alle benoetigten GPIO Pins
   -------------------------------------------------------- */
void seg7_initall(void)
{
  kseg0_init(); kseg1_init(); kseg2_init();
  kseg3_init(); kseg4_init(); kseg5_init();
  kseg6_init(); kseg7_init();

  a_seginit(); b_seginit(); c_seginit();
  d_seginit(); e_seginit(); f_seginit();
  g_seginit(); dp_seginit();
}

/* --------------------------------------------------------
                         key_initall

     initialisiert alle benoetigten GPIO Pins
   -------------------------------------------------------- */
void key_initall(void)
{
  key_0_4_init();
  key_1_5_init();
  key_2_6_init();
  key_3_7_init();

  key_stat= 0;
}


/* --------------------------------------------------------
                         seg7_setbmp

     schaltet einzelne Segmente einer Anzeige an oder aus,
     je nachdem, welches Bitmuster uebergeben wurde;
   -------------------------------------------------------- */
void seg7_setbmp(uint8_t bmp)
{
  if (bmp & 0x01) a_segset();  else a_segclr();
  if (bmp & 0x02) b_segset();  else b_segclr();
  if (bmp & 0x04) c_segset();  else c_segclr();
  if (bmp & 0x08) d_segset();  else d_segclr();
  if (bmp & 0x10) e_segset();  else e_segclr();
  if (bmp & 0x20) f_segset();  else f_segclr();
  if (bmp & 0x40) g_segset();  else g_segclr();
  if (bmp & 0x80) dp_segset(); else dp_segclr();
}

/* --------------------------------------------------------
                         seg7_allclr

     schaltet alle Segmente einer Anzeige aus, funktioniert
     schneller als seg7_setbmp(0)
   -------------------------------------------------------- */
void seg7_allclr(void)
{
  a_segclr();
  b_segclr();
  c_segclr();
  d_segclr();
  e_segclr();
  f_segclr();
  g_segclr();
  dp_segclr();
}

/* --------------------------------------------------------
                         fb_clear

     loescht den Framebuffer (alle LEDs aus)
   -------------------------------------------------------- */
void fb_clear(void)
{
  uint8_t i;

  for (i= 0; i<8; i++) seg7_fb[i]= 0;
}

/* --------------------------------------------------------
                         fb_sethexbyte

     beschreibt den Framebuffer mit einem 2 stelligen
     Hexadezimalwert (8 Bit) ab Position dpos (POS 0
     entspricht rechter Anzeigeposition)
   -------------------------------------------------------- */
void fb_sethexbyte(uint8_t val, uint8_t dpos)
{
  seg7_fb[dpos]   = seg7_bmp[val & 0x0f];
  seg7_fb[dpos+1] = seg7_bmp[(val & 0xf0) >> 4];
}

/* --------------------------------------------------------
                         fb_sethexint

     beschreibt den Framebuffer mit einem 4 stelligen
     Hexadezimalwert (16 Bit) ab Position dpos (POS 0
     entspricht rechter Anzeigeposition)
   -------------------------------------------------------- */
void fb_sethexint(uint16_t val, uint8_t dpos)
{
  seg7_fb[dpos]   = seg7_bmp[val & 0x000f];
  seg7_fb[dpos+1] = seg7_bmp[(val & 0x00f0) >> 4];
  seg7_fb[dpos+2] = seg7_bmp[(val & 0x0f00) >> 8];
  seg7_fb[dpos+3] = seg7_bmp[(val & 0xf000) >> 12];
}

/* --------------------------------------------------------
                         fb_sethexint32

     beschreibt den Framebuffer mit einem 8 stelligen
     Hexadezimalwert (32 Bit) )
   -------------------------------------------------------- */
void fb_sethexint32(uint32_t val)
{
  uint16_t z;

  z= val & 0x0000ffff;
  fb_sethexint(z,0);
  z= val >> 16;
  fb_sethexint(z,4);
}

/* --------------------------------------------------------
                         fb_setdez2

     beschreibt den Framebuffer mit einem 2 stelligen
     Dezimalwert (0..99) ab Position dpos (POS 0
     entspricht rechter Anzeigeposition)
   -------------------------------------------------------- */
void fb_setdez2(uint8_t val, uint8_t dpos)
{
  uint8_t i;

  seg7_fb[dpos] = seg7_bmp[val % 10];
  seg7_fb[dpos+1] = seg7_bmp[val / 10];

}

/* --------------------------------------------------------
                         fb_setdezint

     beschreibt den Framebuffer mit einem 4 stelligen
     Dezimalwert (16 Bit) ab Position dpos (POS 0
     entspricht rechter Anzeigeposition)
   -------------------------------------------------------- */
void fb_setdezint(uint16_t val, uint8_t dpos)
{
  uint8_t i;

  for (i= 0; i< 4; i++)
  {
    seg7_fb[dpos+i] = seg7_bmp[val % 10];
    val= val / 10;
  }
}

/* --------------------------------------------------------
                         fb_setdezint32

     beschreibt den Framebuffer mit einem 8 stelligen
     Dezimalwert (32 Bit)
   -------------------------------------------------------- */
void fb_setdezint32(uint32_t val)
{
  uint8_t i;

  for (i= 0; i< 8; i++)
  {
    seg7_fb[i] = seg7_bmp[val % 10];
    val= val / 10;
  }
}



/* --------------------------------------------------------
                     Interruptvector Timer1

     uebernimmt die Frequenzzaehlervariable in die Variable
     die angezeigt wird und setzt die Zaehlervariable
     zurueck
   -------------------------------------------------------- */
ISR (TIMER1_COMPA_vect)
{
  static   uint8_t gk_cnt = 0;
  static   uint8_t pwmarr[8] = {0,0,0,0,0,0,0,0};

  volatile uint8_t bmp_tmp;


  seg7_allclr();

  switch (gk_cnt)
  {
    case 0 : kseg0_set(); break;
    case 1 : kseg1_set(); break;
    case 2 : kseg2_set(); break;
    case 3 : kseg3_set(); break;
    case 4 : kseg4_set(); break;
    case 5 : kseg5_set(); break;
    case 6 : kseg6_set(); break;
    case 7 : kseg7_set(); break;
    default : break;
  }

  gk_cnt++;
  gk_cnt = gk_cnt % 8;

  switch (gk_cnt)
  {
    case 0 :
    {
      kseg0_clr();

      // Keyboardbank 0 einlesen
      asm volatile ("nop");           // Wartezyklen, damit die MPX-Leitung auch garantiert 0 ist
      asm volatile ("nop");

      if (is_key_0_4()) key_stat &= ~(0x01); else key_stat |= 0x01;
      if (is_key_1_5()) key_stat &= ~(0x02); else key_stat |= 0x02;
      if (is_key_2_6()) key_stat &= ~(0x04); else key_stat |= 0x04;
      if (is_key_3_7()) key_stat &= ~(0x08); else key_stat |= 0x08;

      break;
    }
    case 1 :
    {
      kseg1_clr();

      // Keyboardbank 1 einlesen
      asm volatile ("nop");           // Wartezyklen, damit die MPX-Leitung auch garantiert 0 ist
      asm volatile ("nop");

      if (is_key_0_4()) key_stat &= ~(0x10); else key_stat |= 0x10;
      if (is_key_1_5()) key_stat &= ~(0x20); else key_stat |= 0x20;
      if (is_key_2_6()) key_stat &= ~(0x40); else key_stat |= 0x40;
      if (is_key_3_7()) key_stat &= ~(0x80); else key_stat |= 0x80;

      break;
    }
    case 2 : kseg2_clr(); break;
    case 3 : kseg3_clr(); break;
    case 4 : kseg4_clr(); break;
    case 5 : kseg5_clr(); break;
    case 6 : kseg6_clr(); break;
    case 7 : kseg7_clr(); break;
    default : break;
  }


  // Bitmapmuster des 7-Segmentdigits in Abhaengigkeit des
  // aktuell aktiven Segments laden
  bmp_tmp = seg7_fb[gk_cnt];
  if (seg7_dpmask & (1 << gk_cnt)) bmp_tmp |= 0x80;    // Dezimalpunktanzeige


  // Software-PWM auf einzelne Digits anwenden
  if (dig_pwm[gk_cnt]== 1)
  {
    seg7_setbmp(bmp_tmp);
  }
  else
  {
    if (pwmarr[gk_cnt] == dig_pwm[gk_cnt])
    {
      seg7_setbmp(bmp_tmp);
      pwmarr[gk_cnt]= 0;
    }
    else
    {
      pwmarr[gk_cnt]++;
      pwmarr[gk_cnt] %= 17;
    }
  }
}


/* --------------------------------------------------------
                           mpx8_init

     initialisiert Timer1 Compare Interrupt Modus sowie
     alle am Multiplexing beteiligten I/O Pins
   -------------------------------------------------------- */
void mpx8_init(void)
{
  /* ---------------------------------------------
     CS10 / CS11 / CS12 sind Steuerbits
     des Taktvorteilers.

     CS 12 | CS11 | CS10
     ----------------------------------------------
       0      0      0   kein Takt, Timer gestoppt
       0      0      1   clk / 1 (kein Teiler)
       0      1      0   clk / 8
       0      1      1   clk / 64
       1      0      0   clk / 256
       1      0      1   clk / 1024
     ---------------------------------------------- */
  TCCR1B = 1 << WGM12 | 1 << CS11;       // Taktvorteiler = 8
//  OCR1A = 2500;
  OCR1A = 180;
  TCNT1 = 0;

  TIMSK1 = 1 << OCIE1A;
  sei();
  seg7_initall();
  key_initall();
}
